import numpy as np
import sympy as sy

from sympy import sin, cos


class Example:
    states = ["phi1", "phi1p", "phi2", "phi2p"]

    parameters = {
        "delta": 0.1,
        "gamma": 0.009,
    }

    def differential_system(self):
        return {
            "phi1": phi1p,
            "phi1p": sin(phi1 - gamma),
            "phi2": phi2p,
            "phi2p": sin(phi1 - gamma) + (phi1p ** 2) * sin(phi2) - cos(phi1 - gamma) * sin(phi2),
        }

    def guard(self):
        return round(2 * phi1 - phi2, 5) == 0 and phi2 < -delta

    def reset(self):
        return {
            "phi1": -phi1,
            "phi1p": phi1p * cos(2 * phi1),
            "phi2": -2 * phi1,
            "phi2p": phi1p * cos(2 * phi1) * (1 - cos(2 * phi1))
        }

    system_config = {
        "dt": 0.00002,   # Time steps
        "initial_gap_epsilon": 0.001,
        "w": 0.0001,
        "num_periods": 5,
        "num_system_simulations": 2,
        "auto_period_value": True,
        "period_value": np.nan,
        # If the period value is fixed (auto_period_value==False), give here the value of one period.
        "init_system": [0.009, -0.05869, -0.0009629, -0.3432],
    }

    system_display = {
        "1D_plot": [["phi1"], ["phi1p"], ["phi2"], ["phi2p"]],  # Select states to show as a function of t
        "2D_plot": [["phi1", "phi1p"], ["phi2", "phi2p"]],  # Select two states in each list to show in 2D plot
        "3D_plot": [["phi1", "phi1p", "phi2"]],  # Select three states in each list to show in 3D plot
        ## SOME OPTIONS TO ADD IN 1D PLOTS
        "delta_curve": True,  # To show the curve of delta in green
        "color_depend_on_lambda_sign": False,  # To show the curve of the state according to the sign of lambda (red if lambda>0 and green otherwise)
        "simulations_curves": False,  # To show the curves of the random simulations in blue
        "min_max_tube": True,  # To show the lowest point of the top part of the tube  and the highest point of the bottom part of the tube

        "plot_lambda": True,  # To show the evolution of lambda as a function of t
        "plot_delta": True,  # To show the evolution of delta as a function of t
        "plot_gamma": True,  # To show the evolution of gamma as a function of t
        "step_plot": 1
    }

    def __init__(self, ):
        self.states = self.states
        def_s(self.states)
        try:
            self.parameters = self.parameters
            def_p(self.parameters)
        except AttributeError:
            pass

    def guard_def(self, *states_):
        def_v(self.states, states_)
        if "guard" not in dir(self):
            return None
        else:
            return self.guard()


def def_s(var):
    for i_states in var:
        globals()[i_states] = sy.Symbol(i_states)


def def_v(var, states_):
    i_ = 0
    for i_states in var:
        globals()[i_states] = states_[i_]
        i_ += 1


def def_p(var):
    l_parameters = list(var)
    for i_parameter in l_parameters:
        globals()[i_parameter] = var[i_parameter]
    return globals()
