import numpy as np
import sympy as sy
import math
from sympy import sin, cos


class Example:
    states = ["u1", "v1", "u2", "v2", "u3", "v3", "u4", "v4"]

    parameters = {
        "u0": 1,
        "u5": 1,
        "v0": 3,
        "v5": 3,
        "A": 1,
        "B": 3,
        "sigma": 0.5,
        "l": 1,
    }

    def differential_system(self):
        return {
            "u1": A + u1**2 * v1 -(B+1)*u1 + sigma*(u0-2*u1+u2),
            "v1": B*u1 -u1**2 * v1 + sigma*(v0-2*v1+v2),
			"u2": A + u2**2 * v2 -(B+1)*u2 + sigma*(u1-2*u2+u3),
            "v2": B*u2 -u2**2 * v2 + sigma*(v1-2*v2+v3),
			"u3": A + u3**2 * v3 -(B+1)*u3 + sigma*(u2-2*u3+u4),
            "v3": B*u3 -u3**2 * v3 + sigma*(v2-2*v3+v4),
			"u4": A + u4**2 * v4 -(B+1)*u4 + sigma*(u3-2*u4+u5),
            "v4": B*u4 -u4**2 * v4 + sigma*(v3-2*v4+v5),

        }


    system_config = {
        "dt": 0.002,
        "initial_gap_epsilon": 0.02,
        "w": 0.05,
        "num_periods": 5,
        "num_system_simulations": 2,
        "auto_period_value": True,
        "period_value": np.nan,
        # If the period value is fixed (auto_period_value==False), give here the value of one period.
        "init_system": [1+sin(2*math.pi*(1*1/5)), 3, 1+sin(2*math.pi*(2*1/5)), 3, 1+sin(2*math.pi*(3*1/5)), 3, 1+sin(2*math.pi*(4*1/5)), 3],
    }

    system_display = {
        "1D_plot": [["u1"], ["v1"], ["u2"], ["v2"], ["u3"], ["v3"], ["u4"], ["v4"]],  # Select states to show as a function of t
        "2D_plot": [["u1", "v1"], ["u2", "v2"], ["u3", "v3"], ["u4", "v4"]],  # Select two states in each list to show in 2D plot

        ## SOME OPTIONS TO ADD IN 1D PLOTS
        "delta_curve": True,  # To show the curve of delta in green
        "color_depend_on_lambda_sign": False,  # To show the curve of the state according to the sign of lambda (red if lambda>0 and green otherwise)
        "simulations_curves": False,  # To show the curves of the random simulations in blue
        "min_max_tube": True,  # To show the lowest point of the top part of the tube  and the highest point of the bottom part of the tube

        "plot_lambda": True,  # To show the evolution of lambda as a function of t
        "plot_delta": True,  # To show the evolution of delta as a function of t
        "plot_gamma": True,  # To show the evolution of gamma as a function of t
        "step_plot": 1
    }

    def __init__(self, ):
        self.states = self.states
        def_s(self.states)
        try:
            self.parameters = self.parameters
            def_p(self.parameters)
        except AttributeError:
            pass

    def guard_def(self, *states_):
        def_v(self.states, states_)
        if "guard" not in dir(self):
            return None
        else:
            return self.guard()


def def_s(var):
    for i_states in var:
        globals()[i_states] = sy.Symbol(i_states)


def def_v(var, states_):
    i_ = 0
    for i_states in var:
        globals()[i_states] = states_[i_]
        i_ += 1


def def_p(var):
    l_parameters = list(var)
    for i_parameter in l_parameters:
        globals()[i_parameter] = var[i_parameter]
    return globals()
