import numpy as np
import sympy as sy

from sympy import sin, cos


class Example:
    states = ["r", "y", "theta"]

    parameters = {
        "lambda1": -1.1, 
        "lambda2": 0.1,
        "omega": 30,
    }

    def differential_system(self):
        return {
            "r": y*(lambda1*sin(omega*theta)*cos(omega*theta) - lambda2*sin(omega*theta)*cos(omega*theta)) + (r - 1)*(lambda1*cos(omega*theta)**2 + lambda2*sin(omega*theta)**2),
            "y": y*(lambda1*sin(omega*theta)**2 + lambda2*cos(omega*theta)**2) + (r - 1)*(lambda1*sin(omega*theta)*cos(omega*theta) - lambda2*sin(omega*theta)*cos(omega*theta)),
            "theta": 0.01*sin(theta)**2 + 0.01,
        }
    
    def guard(self):
        return theta >= 2*sy.pi 

    def reset(self):
        return {
            "r": r,
            "y": y,
            "theta": 0,
        }

    system_config = {
        "dt": 0.01,
        "initial_gap_epsilon": 0.2,
        "w": 0.01,
        "num_periods": 4,
        "num_system_simulations": 2,
        "auto_period_value": False,
        "period_value": 444.3,
        # If the period value is fixed (auto_period_value==False), give here the value of one period.
        "init_system": [1.1, 0.1, 0],
    }

    system_display = {
        "1D_plot": [["r"], ["y"], ["theta"]],  # Select states to show as a function of t
        "3D_plot": [["r", "y", "theta"]],  # Select three states in each list to show in 3D plot
        ## SOME OPTIONS TO ADD IN 1D PLOTS
        "delta_curve": True,  # To show the curve of delta in green
        "color_depend_on_lambda_sign": False,  # To show the curve of the state according to the sign of lambda (red if lambda>0 and green otherwise)
        "simulations_curves": False,  # To show the curves of the random simulations in blue
        "min_max_tube": False,  # To show the lowest point of the top part of the tube  and the highest point of the bottom part of the tube

        "plot_lambda": True,  # To show the evolution of lambda as a function of t
        "plot_delta": True,  # To show the evolution of delta as a function of t
        "plot_gamma": True,  # To show the evolution of gamma as a function of t
        "step_plot": 1
    }

    def __init__(self, ):
        self.states = self.states
        def_s(self.states)
        try:
            self.parameters = self.parameters
            def_p(self.parameters)
        except AttributeError:
            pass

    def guard_def(self, *states_):
        def_v(self.states, states_)
        if "guard" not in dir(self):
            return None
        else:
            return self.guard()


def def_s(var):
    for i_states in var:
        globals()[i_states] = sy.Symbol(i_states)


def def_v(var, states_):
    i_ = 0
    for i_states in var:
        globals()[i_states] = states_[i_]
        i_ += 1


def def_p(var):
    l_parameters = list(var)
    for i_parameter in l_parameters:
        globals()[i_parameter] = var[i_parameter]
    return globals()
