import sympy as sy
import numpy as np
from numpy.linalg import norm
import math
from random import uniform
from scipy.optimize import minimize

# def norm(sys_):
#     nor_ = 0
#     if isinstance(sys_, float):
#         return abs(sys_)
#     for i_sys_ in sys_:
#         nor_ += i_sys_**2
#     return sy.sqrt(nor_)
def system_eq_func(f_, symbs_):
    funcs = []
    for i_f in f_:
        funcs.append(sy.lambdify(symbs_, i_f))
    return funcs


def system_eq_func_variables(f_, symbs_):
    funcs = []
    for i_f in f_:
        funcs.append(sy.lambdify(symbs_, i_f))
    return funcs


def deriv_eq(f_, symb_):
    return f_.diff(symb_)


def deriv_eq_variables(f_, symbs_):
    der_eq = []
    for i_syb in symbs_:
        der_eq.append(deriv_eq(f_, i_syb))
    return der_eq


def deriv_sys_variables(f_, symbs_):
    der_sys = []
    for i_f in f_:
        der_sys.append(deriv_eq_variables(i_f, symbs_))
    return der_sys


def get_symbols_from_sys(symbs_):
    return symbs_


def Jacobian_matrix(f_, symbs_):
    jac_mat = []
    symbols_ = get_symbols_from_sys(symbs_)
    d_s_v = deriv_sys_variables(f_, symbols_)
    print("Jacobian_matrix=", d_s_v)
    for i_s in d_s_v:
        i_m = system_eq_func_variables(i_s, symbols_)
        jac_mat.append(i_m)
    return jac_mat


def Jacobian_matrix_values(jac_, symb_values):
    jac_mat_v = []
    # d_s_v = deriv_sys_variables(f_, symbols_)
    for i_s in jac_:
        i_jac_mat_v = []
        for j_s in i_s:
            val_m_i_j = j_s(*symb_values)
            i_jac_mat_v.append(val_m_i_j)
        jac_mat_v.append(i_jac_mat_v)
    return jac_mat_v


def lambda_h(jac_, system_):
    J_p = Jacobian_matrix_values(jac_, system_)
    # print("system_=", type(system_[0]))
    norm_q = 1 / norm(system_) ** 2
    # print("norm_q=", norm_q * (np.dot(system_, J_p).dot(np.array(system_).T)))
    return norm_q * (np.dot(system_, J_p).dot(np.array(system_).T))


def euler_d_system(X_, s_, w_, dt_, guard_equations_, reset_equations_):
    dX_ = []
    for index_s in range(0, len(s_)):
        dX_.append(X_[index_s] + dt_ * (s_[index_s](*X_) + w_))
    # if guard_equations_:
    # print("dX_=", dX_)
    if guard_equations_(*X_):
        # print("g_=")
        # print("dX_0=", dX_)
        for index_s in range(0, len(s_)):
            if reset_equations_[index_s](*X_) is not None:
                dX_[index_s] = reset_equations_[index_s](*X_)
        # print("dX_1=", dX_)
    return dX_


def euler_f_system(X_, s_, w_, guard_equations_, reset_equations_):
    dX_ = []
    for index_s in range(0, len(s_)):
        dX_.append(s_[index_s](*X_) + w_)
    if guard_equations_(*X_):
        for index_s in range(0, len(s_)):
            if reset_equations_[index_s](*X_) is not None:
                dX_[index_s] = reset_equations_[index_s](*X_)
    return dX_


def euler_f_system_no_guard(X_, s_, w_):
    dX_ = []
    for index_s in range(0, len(s_)):
        dX_.append(s_[index_s](*X_) + w_)
    return dX_


def euler_d_system_no_guard(X_, s_, w_, dt_):
    dX_ = []
    for index_s in range(0, len(s_)):
        dX_.append(X_[index_s] + dt_ * (s_[index_s](*X_) + w_))
    return dX_


def dis_2_pts_1D(point1, point2):
    return abs(point1 - point2)


def gamma_u(system_1, system_2, d_system_1, d_system_2, lambda_u, w_):
    diff_d_system = [x - y for x, y in zip(d_system_1, d_system_2)]
    diff_system = [x - y for x, y in zip(system_1, system_2)]
    diff_w = 2 * w_

    # print("norm(diff_w)=", norm(diff_w))
    t1 = diff_d_system
    t2 = diff_system
    a1 = np.dot(t1, t2)
    # print("diff_system=", diff_system, "diff_d_system=", diff_d_system, "scalair product=", a1, "  lambda_u*(norm(diff_system)**2)=", lambda_u*(norm(diff_system)**2), "  norm(diff_system)=", norm(diff_system), "   norm(diff_w)=", norm(diff_w))
    return (a1 - lambda_u * (norm(diff_system) ** 2)) / (norm(diff_system) * norm(diff_w))


def all_gamma_u(all_system_, all_system_next, lambda_u, w_):
    sup_gamma_u = -500000
    # print("all_system_=", all_system_, "all_system_next=", all_system_next, "lambda=", lambda_u)
    # print("len(all_system_)=", len(all_system_), "len(all_system_next)=", len(all_system_next), "lambda=", lambda_u)
    for i_all in range(0, len(all_system_)):
        for j_all in range(0, len(all_system_)):
            # print(" all_w[i_all] != all_w[j_all]", "i_all = ", i_all, "j_all", j_all)
            if i_all != j_all:
                d_system_i = all_system_next[i_all]
                d_system_j = all_system_next[j_all]
                gamma_u_cur = gamma_u(all_system_[i_all], all_system_[j_all], d_system_i,
                                      d_system_j, lambda_u, w_)
                # print("gamma_u_cur=", gamma_u_cur, "   j_all = ", j_all, "        sup_gamma_u=", sup_gamma_u)
                if sup_gamma_u < gamma_u_cur:
                    sup_gamma_u = gamma_u_cur
                    # print("sup_gamma_u=", sup_gamma_u, "   j_all = ", j_all)
    return sup_gamma_u


def Lu(system_1, system_2, d_system_1, d_system_2):
    diff_d_bio = [x - y for x, y in zip(d_system_1, d_system_2)]
    diff_bio = [x - y for x, y in zip(system_1, system_2)]
    return norm(diff_d_bio) / norm(diff_bio)
    # return math.sqrt((d_system_1[0]-d_system_2[0])**2 + (d_system_1[1]-d_system_2[1])**2 + (d_system_1[2]-d_system_2[2])**2) / math.sqrt((system_1[0]-system_2[0])**2 + (system_1[1]-system_2[1])**2 + (system_1[2]-system_2[2])**2)


def Lu_all(all_system_):
    sup_Lu = -5000000
    for i_all in range(0, len(all_system_) - 1):
        for j_all in range(0, len(all_system_) - 1):
            if i_all != j_all:
                d_system_i = all_system_[i_all + 1]
                d_system_j = all_system_[j_all + 1]
                # print(all_system_[i_all], all_system_[j_all], d_system_i, d_system_j)
                Lu_cur = Lu(all_system_[i_all], all_system_[j_all], d_system_i, d_system_j)
                # print(Lu_cur)
                if sup_Lu < Lu_cur:
                    sup_Lu = Lu_cur
    return sup_Lu


def Cu_all(Lu_, all_system_):
    sup_Cu = -5000000
    for i_all in range(1, len(all_system_)):
        Cu_cur = Lu_ * norm(all_system_[i_all])
        if Cu_cur > sup_Cu:
            sup_Cu = Cu_cur
    return sup_Cu


def delta_t_with_uncertainty_lambda_neg(Cu_, lambda_u_, gamma_u_, epsilon_, W_, t_):
    # print(norm(W_))
    l1 = (Cu_ ** 2 / (-lambda_u_ ** 4)) * (
            -(lambda_u_ ** 2) * (t_ ** 2) - 2 * lambda_u_ * t_ + 2 * math.exp(lambda_u_ * t_) - 2)
    l2 = (1 / lambda_u_ ** 2) * (
            ((Cu_ * gamma_u_ * (norm(W_))) / (-lambda_u_)) * (-lambda_u_ * t_ + math.exp(lambda_u_ * t_) - 1))
    l3 = (1 / lambda_u_) * (((gamma_u_ ** 2 * (norm(W_) / 2) ** 2) / (-lambda_u_)) * (math.exp(lambda_u_ * t_) - 1) + (
            lambda_u_ * (epsilon_ ** 2) * math.exp(lambda_u_ * t_)))
    return math.sqrt(l1 + l2 + l3)


def delta_t_without_uncertainty_lambda_neg(Cu_, lambda_u_, epsilon_, t_):
    l1 = (Cu_ ** 2 / (-lambda_u_ ** 4)) * (
            -lambda_u_ ** 2 * t_ ** 2 - 2 * lambda_u_ * t_ + 2 * math.exp(lambda_u_ * t_) - 2)
    # l2 = (1/lambda_u_**2)*(((Cu_*gamma_u_*norm(W_))/(-lambda_u_)) * (-lambda_u_*t_ + math.exp(lambda_u_*t_) - 1))
    l3 = (1 / lambda_u_) * (lambda_u_ * epsilon_ ** 2 * math.exp(lambda_u_ * t_))
    return math.sqrt(l1 + l3)


def delta_t_without_uncertainty_lambda_pos(Cu_, lambda_u_, epsilon_, t_):
    l1 = (epsilon_ ** 2) * math.exp(3 * lambda_u_ * t_)
    l2 = (Cu_ ** 2 / (3 * lambda_u_ ** 2)) * (
            -t_ ** 2 - (2 * t_ / (3 * lambda_u_)) + (2 / (9 * lambda_u_ ** 2)) * (math.exp(3 * lambda_u_ * t_) - 1))
    return math.sqrt(l1 + l2)


def delta_t_with_uncertainty_lambda_pos(Cu_, lambda_u_, gamma_u_, epsilon_, W_, t_):
    coef_ = 1 / ((3 * lambda_u_) ** (3 / 2))
    l1 = (Cu_ ** 2 / lambda_u_) * (
            -9 * (lambda_u_ ** 2) * (t_ ** 2) - 6 * lambda_u_ * t_ + 2 * math.exp(3 * lambda_u_ * t_) - 2)
    l2 = 3 * lambda_u_ * (
            ((Cu_ * gamma_u_ * norm(W_)) / lambda_u_) * (-3 * lambda_u_ * t_ + math.exp(3 * lambda_u_ * t_) - 1))
    l3 = (9 * lambda_u_ ** 2) * (
            (((gamma_u_ ** 2) * ((norm(W_) / 2) ** 2)) / lambda_u_) * (math.exp(3 * lambda_u_ * t_) - 1) + (
            3 * lambda_u_ * (epsilon_ ** 2) * math.exp(3 * lambda_u_ * t_)))
    # print("Cu=", Cu_, "lambda_u", lambda_u_, "gamma_u_=", gamma_u_, "l1=", l1, "l2=", l2, "l3=", l3, "sum_ls=", l1+l2+l3)
    return coef_ * math.sqrt(l1 + l2 + l3)


def delta_t_with_uncertainty_lambda_null(Cu_, gamma_u_, epsilon_, W_, t_):
    l1_ = (Cu_ ** 2) * (- (t_ ** 2) - 2 * t_ + 2 * math.exp(t_) - 2)
    l2_ = Cu_ * gamma_u_ * abs(W_) * (-t_ + math.exp(t_) - 1) + (
                (gamma_u_ ** 2) * ((abs(W_) / 2) ** 2) * (math.exp(t_) - 1) + epsilon_ ** 2 * math.exp(t_))
    # d_ = ((Cu_**2)*(-(t_)**2 - 2*t_ + 2*math.exp(t_) - 2) + (Cu_ * gamma_u_*abs(W_)* (-t_ + math.exp(t_)-1 ) + (gamma_u_**2 * (abs(W_) / 2) ** 2 * (math.exp(t_) - 1) + epsilon_ ** 2 * math.exp(t_))))**(1 / 2)
    return math.sqrt(l1_ + l2_)


def delta_t_without_uncertainty_lambda_null(Cu_, epsilon_, t_):
    l1_ = (Cu_ ** 2) * (- (t_ ** 2) - 2 * t_ + 2 * math.exp(t_) - 2)
    l2_ = epsilon_ ** 2 * math.exp(t_)
    # d_ = ((Cu_**2)*(-(t_)**2 - 2*t_ + 2*math.exp(t_) - 2) + (Cu_ * gamma_u_*abs(W_)* (-t_ + math.exp(t_)-1 ) + (gamma_u_**2 * (abs(W_) / 2) ** 2 * (math.exp(t_) - 1) + epsilon_ ** 2 * math.exp(t_))))**(1 / 2)
    return math.sqrt(l1_ + l2_)


def get_states_name_string(states_):
    return [str(states_i_).replace("_sy", "") for states_i_ in states_]


def index_sys_lambda_sign_change(system_, lambda_):
    index_sys = [0]
    first_sign_positive = True if lambda_[0] > 0 else False
    for i_com in range(1, len(system_)):
        if lambda_[i_com - 1] * lambda_[i_com] < 0:
            index_sys.append(i_com)
    index_sys.append(len(system_) - 1)
    return index_sys, first_sign_positive


def find_period(init_system_, num_periods_, system_equations_, dt_, guard_equations_, reset_equations_):
    prev_prev_system_period = init_system_
    if guard_equations_ is None:
        prev_system_period = euler_d_system_no_guard(prev_prev_system_period, system_equations_, 0, dt_)
        d_system_period = euler_d_system_no_guard(prev_system_period, system_equations_, 0, dt_)
    else:
        prev_system_period = euler_d_system(prev_prev_system_period, system_equations_, 0, dt_, guard_equations_, reset_equations_)
        d_system_period = euler_d_system(prev_system_period, system_equations_, 0, dt_, guard_equations_, reset_equations_)
    i_period = 0
    i_period_1 = []
    while len(i_period_1) <= num_periods_:
        cond_priodicity = (prev_prev_system_period[1] < prev_system_period[1]) and (
                    prev_system_period[1] > d_system_period[1])
        if cond_priodicity:
            # print(i_period)
            i_period_1.append(i_period + 1)

        prev_prev_system_period = prev_system_period
        prev_system_period = d_system_period
        if guard_equations_ is None:
            d_system_period = euler_d_system_no_guard(d_system_period, system_equations_, 0, dt_)
        else:
            d_system_period = euler_d_system(d_system_period, system_equations_, 0, dt_, guard_equations_, reset_equations_)
        i_period += 1
    i_one_period_ = i_period_1[num_periods_ - 1] - i_period_1[num_periods_ - 2] + 1
    return i_one_period_, d_system_period


def system_tubes(NodeT_, length_system_, system_equations_, init_system_, num_system_simu_, guard_equations_, reset_equations_, w_, epsilon_, dt_):
    num_system_tube = num_system_simu_
    system_tube = list(np.zeros((num_system_tube, NodeT_, length_system_)))
    limit_system_top = [x + epsilon_ for x in init_system_]
    limit_system_bottom = [x - epsilon_ for x in init_system_]
    system_tube[0][0] = limit_system_bottom
    system_tube[1][0] = limit_system_top
    init_system_tubes = []
    list_ecart_epsilon = []
    list_w_tube = []
    for i_graph in range(0, num_system_tube):
        w_i = []
        for i in range(0, NodeT_):
            w_i.append(uniform(-w_, w_))
        list_w_tube.append(w_i)
    for i_bios_tube in range(0, num_system_tube - 2):
        eca_rand = uniform(-epsilon_, epsilon_)
        init_system_tubes.append([x + eca_rand for x in init_system_])
        list_ecart_epsilon.append(eca_rand)

    for i_init_bios in range(2, num_system_tube):
        system_tube[i_init_bios][0] = init_system_tubes[i_init_bios - 2]

    for i_system_tube in range(0, num_system_tube):
        for i in range(1, NodeT_):
            system_tube[i_system_tube][i] = euler_d_system(system_tube[i_system_tube][i - 1], system_equations_, list_w_tube[i_system_tube][i - 1], dt_, guard_equations_, reset_equations_)
        # print("end  tube", i_system_tube)
    return system_tube


def objfunc(x_, eqs_, guard_eq_, reset_eq_):
    lc_y1 = x_[:int(len(x_) / 2)]
    lc_y2 = x_[int(len(x_) / 2):]
    if guard_eq_ is None:
        eu_van_y1 = euler_f_system_no_guard(lc_y1, eqs_, 0)
        eu_van_y2 = euler_f_system_no_guard(lc_y2, eqs_, 0)
    else:
        eu_van_y1 = euler_f_system(lc_y1, eqs_, 0, guard_eq_, reset_eq_)
        eu_van_y2 = euler_f_system(lc_y2, eqs_, 0, guard_eq_, reset_eq_)
    diff_d_ = [x - y for x, y in zip(eu_van_y1, eu_van_y2)]
    diff_ = [x - y for x, y in zip(lc_y1, lc_y2)]
    norm_diff_d = norm(diff_d_)
    norm_diff_ = norm(diff_)
    eq_tomax = norm_diff_d / norm_diff_
    f = -eq_tomax
    return f


def constraint1(x_):
    lc_y1 = x_[:int(len(x_) / 2)]
    lc_y2 = x_[int(len(x_) / 2):]
    diff_y1_y2 = [y1_ - y2_ for y1_, y2_ in zip(lc_y1, lc_y2)]
    norm_diff_y1_y2 = norm(diff_y1_y2)
    return norm_diff_y1_y2 - 0.0000001


def lipschitz_constant(eul, eul_next, eqs_, guard_eq_, reset_eq_):
    x0 = np.zeros(len(eul) * 2)
    for i_r in range(0, len(eul)):
        x0[i_r] = eul[i_r]
        x0[len(eul) + i_r] = eul_next[i_r]
    list_dis = []
    bnds = []
    for i_r in range(0, len(eul)):
        list_dis.append(abs(eul[i_r] - eul_next[i_r]))
        bnds.append(tuple([eul[i_r] - list_dis[-1], eul[i_r] + list_dis[-1]]))
    bnds = 2 * bnds
    bnds = tuple(bnds)
    con1 = {'type': 'ineq', 'fun': constraint1}
    cons = ([con1])
    solution = minimize(lambda f_i, eqs_1=eqs_, guard_eq_1=guard_eq_, reset_eq_1=reset_eq_: objfunc(f_i, eqs_1, guard_eq_1, reset_eq_1), x0, method='SLSQP', bounds=bnds, constraints=cons)
    x = solution.x
    res_ = -objfunc(x, eqs_, guard_eq_, reset_eq_)
    return res_


def list_lipschitz_constant(euler_p, eqs_, guard_eq_, reset_eq_):
    res_ = []
    for i_e in range(0, len(euler_p) - 1):
        res_.append(lipschitz_constant(euler_p[i_e], euler_p[i_e + 1], eqs_, guard_eq_, reset_eq_))
    return res_


def list_C_constant(lu_, euler_p, eqs_, guard_eq_, reset_eq_):
    res_ = []
    for i_e in range(0, len(euler_p) - 1):
        res_.append(C_lipschitz(lu_[i_e], euler_p[i_e], euler_p[i_e + 1], eqs_, guard_eq_, reset_eq_))
    return res_


def objfunc_C(x_, l_cons_, eqs_, guard_eq_, reset_eq_):
    lc_y1 = x_
    if guard_eq_ is None:
        eu_van_y1 = euler_f_system_no_guard(lc_y1, eqs_, 0)
    else:
        eu_van_y1 = euler_f_system(lc_y1, eqs_, 0, guard_eq_, reset_eq_)
    norm_f = norm(eu_van_y1)
    eq_tomax = l_cons_ * norm_f
    f = -eq_tomax
    return f


def C_lipschitz(lu_, eul, eul_next, eqs_, guard_eq_, reset_eq_):
    x0 = eul
    list_dis = []
    bnds = []
    for i_r in range(0, len(eul)):
        list_dis.append(abs(eul[i_r] - eul_next[i_r])/2)
        bnds.append(tuple([eul[i_r] - list_dis[-1], eul[i_r] + list_dis[-1]]))
    bnds = tuple(bnds)
    solution = minimize(lambda f_i, l_cons_=lu_, eqs_1=eqs_, guard_eq_1=guard_eq_, reset_eq_1=reset_eq_: objfunc_C(f_i, l_cons_, eqs_1, guard_eq_1, reset_eq_1), x0, method='SLSQP', bounds=bnds)
    x = solution.x
    res_ = -objfunc_C(x, lu_, eqs_, guard_eq_, reset_eq_)
    return res_


def constraint1_gamma(x_):
    lc_y1 = x_[:int((len(x_)-2) / 2)]
    lc_y2 = x_[int((len(x_)-2) / 2):-2]
    diff_y1_y2 = [y1_ - y2_ for y1_, y2_ in zip(lc_y1, lc_y2)]
    norm_diff_y1_y2 = norm(diff_y1_y2)
    # print("cont=", norm_diff_y1_y2 - 0.0001)
    return norm_diff_y1_y2 - 0.00000001


def constraint2_gamma(x_):
    w1_ = x_[-2]
    w2_ = x_[-1]
    return w1_ - w2_ - 0.0001


def constraint3_gamma(x_):
    w1_ = x_[-2]
    w2_ = x_[-1]
    return w2_ - w1_ - 0.0001


def objfunc_gamma(x_, lambda_u, eqs_, guard_eq_, reset_eq_):
    lc_y1 = x_[:int((len(x_)-2) / 2)]
    lc_y2 = x_[int((len(x_)-2) / 2):-2]
    w1_ = x_[-1]
    w2_ = x_[-2]
    if guard_eq_ is None:
        eu_van_y1 = euler_f_system_no_guard(lc_y1, eqs_, w1_)
        eu_van_y2 = euler_f_system_no_guard(lc_y2, eqs_, w2_)
    else:
        eu_van_y1 = euler_f_system(lc_y1, eqs_, w1_, guard_eq_, reset_eq_)
        eu_van_y2 = euler_f_system(lc_y2, eqs_, w2_, guard_eq_, reset_eq_)
    eq_tomax = gamma_u(lc_y1, lc_y2, eu_van_y1, eu_van_y2, lambda_u, w1_*2)
    f = -eq_tomax
    return f


def gamma_lipschitz(eul, eul_next, lambda_, eul_max, eul_min, eqs_, guard_eq_, reset_eq_, w_):
    # x0 = np.zeros(len(eul) * 2 + 2)
    x0 = [0]*(len(eul) * 2 + 2)
    for i_r in range(0, len(eul)):
        x0[i_r] = eul[i_r]
        x0[len(eul) + i_r] = eul_next[i_r]
    x0[-2] = -w_/2
    x0[-1] = w_/2
    list_dis = []
    bnds = []
    for i_r in range(0, len(eul)):
        d1_ = abs(eul[i_r] - eul_next[i_r])
        # d3_ = abs(eul[i_r] - eul_min[i_r])
        # d4_ = abs(eul_max[i_r] - eul[i_r])
        # list_dis.append(max(d1_, d3_, d4_))
        list_dis.append(d1_)
        bnds.append(tuple([eul[i_r] - list_dis[-1], eul[i_r] + list_dis[-1]]))
    bnds = 2 * bnds
    bnds.append(tuple([-w_/2, w_/2]))
    bnds.append(tuple([-w_/2, w_/2]))
    bnds = tuple(bnds)
    # print("bnds=", bnds)
    # print("x0=", x0)
    con1_gamma = {'type': 'ineq', 'fun': constraint1_gamma}
    con2_gamma = {'type': 'ineq', 'fun': constraint2_gamma}
    con3_gamma = {'type': 'ineq', 'fun': constraint3_gamma}
    cons_gamma = [con1_gamma, con2_gamma, con3_gamma]
    solution = minimize(lambda f_i, lam=lambda_, eqs_1=eqs_, guard_eq_1=guard_eq_, reset_eq_1=reset_eq_: objfunc_gamma(f_i, lam, eqs_1, guard_eq_1, reset_eq_1), x0, method='SLSQP', bounds=bnds,
                        constraints=cons_gamma)
    x = solution.x
    res_ = -objfunc_gamma(x, lambda_, eqs_, guard_eq_, reset_eq_)
    return res_


def gamma_delta_wall(system_orig_, system_max_, system_min_, all_lambda_, NodeT_, length_system_, system_equations_, w_, guard_equations_, reset_equations_, epsilon_, dt_):
    # step_gamma = int(i_one_period / 20000) + 2
    step_gamma = 2
    saut = int(step_gamma / 2)
    step_gamma = saut * 2
    # print("step_gamma=", step_gamma)
    # print("jump_gamma=", saut)
    all_gamma = []
    wall_tube_top = list(np.zeros((NodeT_, length_system_)))
    wall_tube_bottom = list(np.zeros((NodeT_, length_system_)))
    all_delta = list(np.zeros(NodeT_))
    all_delta[0] = epsilon_
    wall_tube_top[0] = [x + epsilon_ for x in system_orig_[0]]
    wall_tube_bottom[0] = [x - epsilon_ for x in system_orig_[0]]
    system_gamma_orig = list(np.zeros((NodeT_ + step_gamma, length_system_)))
    system_gamma_orig[:NodeT_] = system_orig_[:NodeT_]
    if guard_equations_ is None:
        for i in range(NodeT_ - 1, NodeT_ + step_gamma):
            system_gamma_orig[i] = euler_d_system_no_guard(system_gamma_orig[i - 1], system_equations_, 0, dt_)
    else:
        for i in range(NodeT_ - 1, NodeT_ + step_gamma):
            system_gamma_orig[i] = euler_d_system(system_gamma_orig[i - 1], system_equations_, 0, dt_, guard_equations_, reset_equations_)
    loading = 0
    Lu = list_lipschitz_constant(system_gamma_orig[:NodeT_ + 1], system_equations_, guard_equations_, reset_equations_)
    Cu = list_C_constant(Lu, system_gamma_orig[:NodeT_ + 1], system_equations_, guard_equations_, reset_equations_)
    for i in range(1, NodeT_):
        if w_ != 0:
            comp = i - 1
            # if i % int(NodeT_ / 100) == 0:
            #     loading += 1
                # print(f'{loading}/100')
            # if all_lambda_[i] < 0:
            gamma_i = all_gamma_u(system_gamma_orig[comp:comp + step_gamma:saut], system_gamma_orig[comp + saut:comp + step_gamma + saut:saut], all_lambda_[i], w_)
            # gamma_i = gamma_lipschitz(system_gamma_orig[comp], system_gamma_orig[comp + 1], all_lambda_[i - 1], system_max_[comp], system_min_[comp],
            #                           system_equations_, guard_equations_, reset_equations_, w_)
            if math.isnan(gamma_i) or gamma_i < 0:
                all_gamma.append(0)
                print(gamma_i)
            else:
                all_gamma.append(gamma_i)
            # else:
            #     all_gamma.append(0)

            # if all_gamma[i - 1] > 0:
            #     all_delta[i] = all_delta[i - 1] * math.exp(all_lambda_[i] * dt_) + (
            #             all_gamma[i - 1] * 2 * abs(w_) * dt_ / math.sqrt(abs(all_lambda_[i])))
            # else:
            #     all_delta[i] = all_delta[i - 1] * math.exp(all_lambda_[i] * dt_)

            if all_lambda_[i] > 0:
                all_delta[i] = delta_t_with_uncertainty_lambda_pos(Cu[i], all_lambda_[i], all_gamma[i - 1], all_delta[i - 1], w_, dt_)
            elif all_lambda_[i] < 0:
                all_delta[i] = delta_t_with_uncertainty_lambda_neg(Cu[i], all_lambda_[i], all_gamma[i - 1], all_delta[i - 1], w_, dt_)
            else:
                all_delta[i] = delta_t_with_uncertainty_lambda_null(Cu[i], all_gamma[i - 1], all_delta[i - 1], w_, dt_)
        # else:
        #     all_delta[i] = all_delta[i - 1] * math.exp(all_lambda_[i] * dt_)

            if all_lambda_[i] > 0:
                all_delta[i] = delta_t_without_uncertainty_lambda_pos(Cu[i], all_lambda_[i], all_delta[i - 1], dt_)
            elif all_lambda_[i] < 0:
                all_delta[i] = delta_t_without_uncertainty_lambda_neg(Cu[i], all_lambda_[i], all_delta[i - 1], dt_)
            else:
                all_delta[i] = delta_t_without_uncertainty_lambda_null(Cu[i], all_delta[i - 1], dt_)

        wall_tube_bottom[i] = [x - all_delta[i] for x in system_orig_[i]]
        wall_tube_top[i] = [x + all_delta[i] for x in system_orig_[i]]

    return all_delta, all_gamma, wall_tube_bottom, wall_tube_top


def inclusion_check(periods_, i_one_period_, wall_tube_bottom_, wall_tube_top_, length_system_, all_lambda_, all_delta_,
                    system_orig_, ecart_epsilon_, states_name_, dt_):
    check_inc = "\nCheck inclusion:\n"
    inclusion_verif_index = []
    for i_periods in range(0, periods_ + 1):
        i_one_period_1 = i_one_period_
        # p1 = "Period:" + str(i_periods) + " Tube_top:" + str(
        #     wall_tube_top_[i_periods * i_one_period_1]) + " Tube_bottom:" + str(
        #     wall_tube_bottom_[i_periods * i_one_period_1])
        # print("i=", i_periods * i_one_period_1, "period:", i_periods, "tube_top:",
        #       wall_tube_top_[i_periods * i_one_period_1], " tube_bottom: ",
        #       wall_tube_bottom_[i_periods * i_one_period_1])
        # check_inc += p1 + "\n"
        list_test_in = []
        p4 = "Y_0(" + str(i_periods) + "T)=" + str(system_orig_[i_periods * i_one_period_1]) + \
             ", delta_w(" + str(i_periods) + "T)=" + str(all_delta_[i_periods * i_one_period_1]) + ", epsilon*e^(lambda*t)=" + \
             str(ecart_epsilon_ * math.exp(all_lambda_[i_periods * i_one_period_1] * (i_periods * i_one_period_1 * dt_)))
        check_inc += p4 + "\n"
        if i_periods > 0:
            for i_len_system in range(0, length_system_):
                if wall_tube_bottom_[(i_periods - 1) * i_one_period_1][i_len_system] <= \
                        wall_tube_top_[i_periods * i_one_period_1][i_len_system] <= \
                        wall_tube_top_[(i_periods - 1) * i_one_period_1][i_len_system] and \
                        wall_tube_bottom_[(i_periods - 1) * i_one_period_1][i_len_system] <= \
                        wall_tube_bottom_[i_periods * i_one_period_1][i_len_system] <= \
                        wall_tube_top_[(i_periods - 1) * i_one_period_1][i_len_system]:
                    p2 = "For state " + str(states_name_[i_len_system]) + ": tube at t=" + str(i_periods) + "T is included in tube at t=" + str(
                        i_periods - 1)+"T"
                    check_inc += p2 + "\n"
                    # print(states_name_[i_len_system], ": period ", i_periods, " in period ", i_periods - 1)
                    list_test_in.append(True)
                else:
                    list_test_in.append(False)

            if all(list_test_in):
                p3 = "For all the states: Tubes at t=" + str(i_periods) + "T are included in tubes at t=" + str(i_periods - 1) + "T."
                check_inc += p3 + "\n"
                inclusion_verif_index.append(True)
                # print("all period ", i_periods, " in period ", i_periods - 1, )
            else:
                inclusion_verif_index.append(False)
    compt_i_ = len(inclusion_verif_index)
    p4 = "Conclusion: "
    # for i_in in reversed(inclusion_verif_index):
    #     if i_in:
    #         p4 += "B("+str(compt_i_) + "T) included in "
    #         compt_i_ -= 1
    #     else:
    #         break
    # if compt_i_ == len(inclusion_verif_index):
    #     p4 += "The inclusion not verified in the last period!\n"
    #     compt_i_ -= 1
    # else:
    #     p4 += "B("+str(compt_i_) + "T)\n"
    check_no_inclu = False
    for i_in in reversed(inclusion_verif_index):
        if i_in:
            p4 += "B("+str(compt_i_) + "T) included in " + "B("+str(compt_i_-1) + "T)\n"
            compt_i_ -= 1
            check_no_inclu = True
    if not check_no_inclu:
        p4 += "There is no inclusion\n"
    check_inc += p4
    return check_inc, compt_i_


def sum_lambda_sign_check(all_lambda_, periods_, i_one_period_):
    check_sum_lam = "\nCheck that the sum of lambda is negative for each period:\n"
    for i in range(0, periods_):
        p = "Period number " + str(i + 1) + ":" + " sum(lambda) = " + str(
            sum(all_lambda_[i * i_one_period_: (i + 1) * i_one_period_]))
        check_sum_lam += p + "\n"
        # print(i + 1, "period: from", i_one_period_ * i, "->", i_one_period_ * (i + 1), "sum(lambda) = ",
        #       sum(all_lambda_[i * i_one_period_: (i + 1) * i_one_period_]))
    return check_sum_lam


def min_wall_top_max_wall_bottom_check(length_system_, i_one_period_, wall_tube_top_, wall_tube_bottom_, i_period_inclus_, states_name_):
    i_min_wall_top = np.zeros(length_system_)
    i_max_wall_bottom = np.zeros(length_system_)
    min_wall_top = 10 * np.ones(length_system_)
    max_wall_bottom = -10 * np.ones(length_system_)
    i_period_inclus_=0
    for i_len_system in range(0, length_system_):
        for i in range(i_one_period_ * i_period_inclus_, i_one_period_ * (i_period_inclus_ + 1) + 1):
            if wall_tube_top_[i][i_len_system] < min_wall_top[i_len_system]:
                min_wall_top[i_len_system] = wall_tube_top_[i][i_len_system]
                i_min_wall_top[i_len_system] = i
            if wall_tube_bottom_[i][i_len_system] > max_wall_bottom[i_len_system]:
                max_wall_bottom[i_len_system] = wall_tube_bottom_[i][i_len_system]
                i_max_wall_bottom[i_len_system] = i
    check_min_max = "\nCheck for at least one stat that the lower limit of the top of the tube is less than the higher limit of the bottom of the tube:\n"
    # print("i_min=", i_min_wall_top, "min_wall_top=", min_wall_top)
    # print("i_max=", i_max_wall_bottom, "max_wall_bottom=", max_wall_bottom)
    checked_min_max = []
    for i_len_system in range(0, length_system_):
        check_min_max += "For the state "+states_name_[i_len_system]+":"+" the minimum value of the top part of the tube =" + str(min_wall_top[i_len_system]) + " , the maximum value of the bottom part of the tube =" + str(max_wall_bottom[i_len_system]) + "\n"
        if min_wall_top[i_len_system] <= max_wall_bottom[i_len_system]:
            checked_min_max.append(states_name_[i_len_system])
    check_min_max += "The proposition is verified for the states: " + str(checked_min_max) + "\n"
    return check_min_max, i_min_wall_top, i_max_wall_bottom, min_wall_top, max_wall_bottom
